; WATCHDOG timer handler, Version 1.0
; Copyright 1988 by Ken Sykes.  All rights reserved.
;
; This program may be used freely as long as the copyright notices are
; preserved and it is not sold for profit.  The author makes no warranties,
; expressed or implied, of the accuracy or fitness of this program.
; 
; DESCRIPTION
;    This program takes advantage of timer channel 3, the watchdog timer,
; to avoid cold boots after a system crash.  The watchdog monitors channel
; 0 (system timer) and decrements its counter if a system clock tick occurs
; while IRQ 0 is active.  In other words, if the system timer interrupt is
; not available to service the tick, the watchdog counter is decremented.
; If the watchdog counter reaches 0, an NMI is triggered.
;    WATCHDOG enables the watchdog timer and provides an NMI routine that
; restores the machine stability to the point where the keyboard will be
; serviced properly.  In particular, software interrupts are enabled and
; the 8259A Mask register is set to enable keyboard interrupts.  Bus 
; arbitration, which was turned off for the NMI, is resumed again.
;
; This timer is available ONLY on PS/2 Models 50 and above.
;
; USING WATCHDOG
;     To install watchdog: watchdog
;   To uninstall watchdog: watchdog u  (must be lowercase 'u')
;   If the system hangs, press CTRL-ALT-DEL as always
;
; KNOWN BUGS
;   - Only a portion of the used memory is reclaimed during an uninstall
;   - Once the NMI is triggered, the watchdog is disabled forever (?)
;     Originally the watchdog was re-enabled on exit from the NMI, but
;     system resets would not work properly.  The code as-is seems to
;     break out of VERY tight loops and out of multiple areas of disabled
;     interrupts.  I would welcome explanations of why this works (or doesn't
;     work) from people who have used this timer.


        .286                            ; we'll be allowed to use 286 ops

StdOut  equ     1                       ; handle for standard output

; handy structure for accessing components of DWORD
HILO    struc
lo      dw      ?
hi      dw      ?
HILO    ends

cseg    segment
        assume cs:cseg

        org     100h                    ; we want this to be a .com program
Start:
        jmp     InitHandler             ; jump to initialization section

; this is where our data is located

OldNMIRoutine   dd      ?               ; address of old NMI handler
ValidNMIRoutine dw      ?               ; is old NMI handler valid? (not 0:0)

Signature       db      'WD 1.0 by Ken Sykes'
SignatureLen    equ     $ - Signature

NMIHandler      proc    far
        sti                             ; allow interrupts
        pushf

        push    ax                      ; check to see if watchdog triggered
        in      al, 92h                 ; the NMI by looking at
        test    al, 00010000b           ; bit 4 of port 92 hex
        pop     ax                      ;
        jnz     WatchDogNMI             ; it's set, so watchdog did it

        cmp     cs:ValidNMIRoutine, 0   ; is old NMI valid?
        jne     GotoOldNMIRoutine       ; yes, transfer to it

        stc                             ; no, set carry flag
        ret     4                       ; and return

GotoOldNMIRoutine:
        push    cs:(OldNMIRoutine.hi)   ; stupid assembler made me do this
        push    cs:(OldNMIRoutine.lo)
        ret     2                       ; should just be a far jump off cs

WatchDogNMI:
        push    ax

        mov     ax, 0c300h              ; disable watchdog timer
        int     15h                     ;

        in      al, 90h                 ; turn bus arbitration back on
        and     al, 10111111b           ;
        out     90h, al                 ;

        cli
        in      al, 21h                 ; make sure kbd not masked out
        and     al, 11111101b           ;
        out     21h, al                 ;
        sti

        pop     ax
        popf
        ret     2
NMIHandler      endp

EndOfResidentCode equ this byte         ; everything below here will be
                                        ; gone when we exit to DOS

SignOnMessage   db      'WATCHDOG timer handler, Version 1.0',13,10
                db      'Copyright 1988 by Ken Sykes.  All rights reserved.',13,10
SignOnMessageLen        equ     $ - SignOnMessage

SuccessMessage  db      13,10,'CTRL-ALT-DEL should break out of ''hard'' crashes now.',13,10
SuccessMessageLen       equ     $ - SuccessMessage

NotPS2Message   db      13,10,'This program requires a PS/2 Model 50 or higher.',13,10
NotPS2MessageLen        equ     $ - NotPS2Message

InstalledMessage        db      13,10,'WATCHDOG already installed.',13,10
InstalledMessageLen     equ     $ - InstalledMessage

UninstalledMessage        db      13,10,'WATCHDOG removed from memory.',13,10
UninstalledMessageLen     equ     $ - UninstalledMessage

; Command line processor
CheckOpts       proc    near
        push    ds
        push    es

        mov     ax, cs
        mov     ds, ax

        mov     ah, 51h                 ; get PSP address
        int     21h
        mov     es, bx                  ;

        mov     di, 80h                 ; point to command length
        mov     cl, byte ptr es:[di]    ; get length
        xor     ch, ch                  ;
        inc     cx
        inc     di                      ; move past length

        mov     al, 'u'                 ; look for uninstall option
        cld                             ;
        repne   scasb                   ;
        jcxz    doneCheckOpts           ; not found

        mov     ax, 0c300h              ; disable watchdog time
        int     15h

        les     bx, OldNMIRoutine         ; determine old NMI address
        add     bx, offset OldNMIRoutine  ;
        sub     bx, 100h                  ;
        mov     dx, es:[bx]               ; save it's contents
        mov     ds, es:[bx+2]             ;

        mov     ax, 2502h               ; restore interrupt vector
        int     21h                     ;
        mov     ah, 49h                 ; free space used by watchdog
        int     21h                     ;

        mov     ax, cs
        mov     ds, ax
        mov     dx, offset UninstalledMessage   ; display signon message
        mov     cx, UninstalledMessageLen       ;
        mov     bx, StdOut                      ;
        mov     ah, 40h                         ;
        int     21h                             ;

        mov     ax, 4c00h               ; exit successfully
        int     21h

doneCheckOpts:
        pop     es
        pop     ds
        ret
CheckOpts       endp

; Initialization code
InitHandler     proc
        mov     ax, cs                          ; ds = cs
        mov     ds, ax                          ;

        mov     dx, offset SignOnMessage        ; display signon message
        mov     cx, SignOnMessageLen            ;
        mov     bx, StdOut                      ;
        mov     ah, 40h                         ;
        int     21h                             ;

        mov     ah, 0c0h                        ; get system descriptor addr.
        int     15h
        jc      NotPS2                          ; if carry set it isn't PS/2
        cmp     byte ptr es:[bx+2], 0f8h        ; is it model 80?
        je      ContinueInit                    ; yes, so hardware ok
        cmp     byte ptr es:[bx+2], 0fch        ; is model # valid for 50/60?
        jne     NotPS2                          ; no, it's not a PS/2
        cmp     byte ptr es:[bx+3], 4           ; is submodel high enough?
        jae     ContinueInit                    ; yup, so it's a 50/60

NotPS2:
        mov     dx, offset NotPS2Message        ; display failure message
        mov     cx, NotPS2MessageLen            ;
        mov     bx, StdOut                      ;
        mov     ah, 40h                         ;
        int     21h                             ;
        
        mov     ax, 4c01h                       ; terminate with failure
        int     21h                             ;

ContinueInit:
        push    es                              ; save PSP segment

        mov     ax, 3502h                       ; get current NMI address
        int     21h                             ;
        mov     ax, es                          ; get segment in register
        mov     OldNMIRoutine.hi, ax            ; save it
        mov     OldNMIRoutine.lo, bx            ;
        or      ax, bx                          ; check for null
        mov     ValidNMIRoutine, ax             ; save flag

        call    CheckOpts                       ; check command line options

        mov     si, offset Signature            ; check for signature
        mov     di, OldNMIRoutine.lo            ;
        mov     cx, SignatureLen
        sub     di, cx
        cld
        repz    cmpsb                           ; compare

        pop     es                              ; restore es

        or      cx, cx                          ; is cx != 0?
        jnz     Install                         ; yes, install

        mov     dx, offset InstalledMessage     ; display installed message
        mov     cx, InstalledMessageLen         ;
        mov     bx, StdOut                      ;
        mov     ah, 40h                         ;
        int     21h                             ;

        mov     ax, 4c02h                       ; exit with error code
        int     21h

Install:
        mov     dx, offset SuccessMessage       ; display success message
        mov     cx, SuccessMessageLen           ;
        mov     bx, StdOut                      ;
        mov     ah, 40h                         ;
        int     21h                             ;

        mov     es, word ptr es:[2ch]           ; free the environment
        mov     ah, 49h                         ;
        int     21h                             ;

        mov     dx, offset NMIHandler           ; get offset
        mov     ax, 2502h                       ; set interrupt
        int     21h

        mov     bx, 18                          ; enable watchdog timer
        mov     ax, 0c301h                      ;
        int     15h                             ;

        mov     dx, offset EndOfResidentCode    ; compute # paragraphs to
        add     dx, 15                          ; save
        shr     dx, 4                           ;
        mov     ax, 3100h                       ; TSR with exit code 0
        int     21h                             ;
InitHandler     endp

cseg    ends
        end     Start




